package provider

import (
	"github.com/Shopify/sarama"
	"log"
	"sync"
)

type ConsumerProvider struct {
	consumersGroupLock    sync.Mutex
	consumersGroup        []sarama.ConsumerGroup
	consumerGroupProvider func() sarama.ConsumerGroup
}

func NewConsumerProvider(brokers []string, groupID string, consumerConfigurationProvider func() *sarama.Config) Provider[sarama.ConsumerGroup] {
	return &ConsumerProvider{
		consumerGroupProvider: func() sarama.ConsumerGroup {
			configuration := consumerConfigurationProvider()

			group, err := sarama.NewConsumerGroup(brokers, groupID, configuration)
			if err != nil {
				log.Fatalln(err)
			}
			return group
		},
	}
}

func (c *ConsumerProvider) Borrow() (consumerGroup sarama.ConsumerGroup) {
	c.consumersGroupLock.Lock()
	defer c.consumersGroupLock.Unlock()

	if len(c.consumersGroup) == 0 {
		for {
			consumerGroup = c.consumerGroupProvider()
			if consumerGroup != nil {
				return
			}
		}
	}

	index := len(c.consumersGroup) - 1
	consumerGroup = c.consumersGroup[index]
	c.consumersGroup = c.consumersGroup[:index]
	return
}

func (c *ConsumerProvider) Release(consumerGroup sarama.ConsumerGroup) {
	c.consumersGroupLock.Lock()
	defer c.consumersGroupLock.Unlock()

	if err := consumerGroup.Close(); err != nil {
		log.Fatalln(err)
	}

	c.consumersGroup = append(c.consumersGroup, consumerGroup)
}

func (c *ConsumerProvider) Clear() {
	c.consumersGroupLock.Lock()
	defer c.consumersGroupLock.Unlock()

	for _, consumer := range c.consumersGroup {
		if err := consumer.Close(); err != nil {
			log.Fatalln(err)
		}
	}
	c.consumersGroup = c.consumersGroup[:0]
}
